@@ -4,35 +4,26 @@
 
 from collections import OrderedDict, defaultdict
 from dataclasses import dataclass
-from enum import Enum, IntEnum, auto
+from enum import Enum, auto
 from types import MappingProxyType
 from typing import Tuple, Dict, Set, Optional
 
-from cuda.tile._ir.type import TupleTy, TokenTy
+from cuda.tile._ir.type import TokenTy
 from cuda.tile._memory_model import MemoryOrder
-from cuda.tile._exception import Loc
-from cuda.tile._ir.ir import Function, Block, IRContext, Var, Operation
+from cuda.tile._exception import Loc, TileInternalError
+from cuda.tile._ir.ir import Block, IRContext, Var, Operation, MemoryEffect
 from cuda.tile._ir.ops import (
-    Assign, Break, BuildTuple, CarriedVariables, Continue, EndBranch, IfElse,
-    IfElseResults, JoinTokens, LoadMemoryOperation, Loop, MakeToken,
-    MemoryOperation, Range, StoreMemoryOperation, TileAtomicCAS, TileAtomicCASTokenOrdered,
+    Break, Continue, EndBranch, IfElse,
+    JoinTokens, Loop, MakeToken,
+    TileAtomicCAS, TileAtomicCASTokenOrdered,
     TileAtomicRMW, TileAtomicRMWTokenOrdered, LoadPointer, LoadPointerTokenOrdered,
     TileLoad, TileLoadTokenOrdered, StorePointer, StorePointerTokenOrdered,
-    TileStore, TileStoreTokenOrdered,
+    TileStore, TileStoreTokenOrdered, TileAssert, TilePrintf,
 )
 from cuda.tile._ir.ops_utils import memory_order_has_acquire, memory_order_has_release
 from cuda.tile._passes.alias_analysis import ALIAS_UNIVERSE, AliasResult, AliasSet
 
 
-class MemoryEffect(IntEnum):
-    # Int value assigned here is meaningful.
-    # It implies the relative strength of memory effects.
-    # For example, NONE < LOAD < STORE.
-    NONE = 0
-    LOAD = 1
-    STORE = 2
-
-
 class MemoryEffects:
     def __init__(self,
                  array_mem_effects: OrderedDict[AliasSet, MemoryEffect],
@@ -87,13 +78,11 @@ TokenKey = AliasTokenKey | AcquireTokenKeyClass
 
 @dataclass(frozen=True)
 class IfElseInfo:
-    ifelse_results: IfElseResults
     ifelse_mem_effects: MemoryEffects
 
 
 @dataclass(frozen=True)
 class InnermostLoopInfo:
-    loop_vars: CarriedVariables
     loop_mem_effects: MemoryEffects
     # TODO: remove once parallel loop store no longer depends on these two
     loop_parallel_stores: Set[Operation]
@@ -103,13 +92,11 @@ class InnermostLoopInfo:
 @dataclass(frozen=True)
 class VarInfo:
     root_var: Dict[str, str]
-    defining_op: Dict[str, Operation]
 
 
 @dataclass(frozen=True)
 class TokenOrderContext:
     alias_result: AliasResult
-    var_info: VarInfo
     block_memory_effects: Dict[Block, MemoryEffects]
 
 
@@ -123,17 +110,16 @@ _TOKEN_ORDERED_OP_MAP = {
 }
 
 
-def token_order_pass(func: Function, alias_result: AliasResult):
+def token_order_pass(root_block: Block, alias_result: AliasResult):
     block_memory_effects = {}
-    _get_block_memory_effects(func.root_block, alias_result, block_memory_effects)
-    var_info = _get_var_info(func.root_block)
-    context = TokenOrderContext(alias_result, var_info, block_memory_effects)
+    _get_block_memory_effects(root_block, alias_result, block_memory_effects)
+    context = TokenOrderContext(alias_result, block_memory_effects)
 
-    root_tok = _make_token_var(func.root_block.ctx, func.loc)
+    root_tok = _make_token_var(root_block.ctx, root_block.loc)
     token_map = defaultdict(lambda: root_tok)
-    _to_token_order_in_block(func.root_block, context, token_map)
+    _to_token_order_in_block(root_block, context, token_map)
     # Ensures Operation.parent_block is correctly set
-    func.root_block[:0] = (MakeToken(root_tok, func.loc),)
+    root_block[:0] = (MakeToken(root_tok, root_block.loc),)
 
 
 def _get_input_var(op: Operation):
@@ -145,20 +131,15 @@ def _get_block_memory_effects(block: Block,
                               block_memory_effects: Dict[Block, MemoryEffects]):
 
     def get_memory_effects(cur_op):
-        if not isinstance(cur_op, MemoryOperation):
+        effect = cur_op.memory_effect
+        if effect == MemoryEffect.NONE or isinstance(cur_op, TileAssert | TilePrintf):
             return EMPTY_MEMORY_EFFECTS
 
-        if isinstance(cur_op, LoadMemoryOperation):
-            effect = MemoryEffect.LOAD
-        else:
-            assert isinstance(cur_op, StoreMemoryOperation)
-            effect = MemoryEffect.STORE
-
         has_acquire_order = False
         if isinstance(cur_op, (TileAtomicCAS, TileAtomicRMW)):
-            has_acquire_order = memory_order_has_acquire(op.memory_order)
+            has_acquire_order = memory_order_has_acquire(cur_op.memory_order)
 
-        return MemoryEffects({alias_result[_get_input_var(op).name]: effect}, has_acquire_order)
+        return MemoryEffects({alias_result[_get_input_var(cur_op).name]: effect}, has_acquire_order)
 
     blk_mem_effects = EMPTY_MEMORY_EFFECTS
     for op in block.operations:
@@ -171,23 +152,6 @@ def _get_block_memory_effects(block: Block,
     block_memory_effects[block] = blk_mem_effects
 
 
-def _get_var_info(root_block: Block) -> VarInfo:
-    defining_op = dict()
-    root_var = dict()
-
-    def traverse(block: Block):
-        for op in block.operations:
-            if isinstance(op, Assign):
-                root_var[op.result_var.name] = root_var.get(op.value.name, op.value.name)
-            for v in op.result_vars:
-                defining_op[v.name] = op
-            for block in op.nested_blocks:
-                traverse(block)
-
-    traverse(root_block)
-    return VarInfo(root_var, defining_op)
-
-
 def _to_token_order_in_block(block: Block,
                              context: TokenOrderContext,
                              token_map: Dict[TokenKey, Var],
@@ -272,18 +236,16 @@ def _to_token_order_in_block(block: Block,
         elif isinstance(op, Loop):
             body_mem_effects = context.block_memory_effects[op.body]
 
-            new_loop_vars = CarriedVariables(
-                list(op.carried_vars.names), list(op.carried_vars.initial),
-                list(op.carried_vars.body), list(op.carried_vars.results)
-            )
+            new_initial_values = list(op.initial_values)
+            new_body_params = list(op.body.params)
+            new_result_vars = list(op.result_vars)
 
             def append_new_carried_var(init_var: Var):
-                new_loop_vars.names.append(init_var.name)
-                new_loop_vars.initial.append(init_var)
+                new_initial_values.append(init_var)
                 body_var = _make_token_var(block.ctx, op.loc)
-                new_loop_vars.body.append(body_var)
+                new_body_params.append(body_var)
                 res_var = _make_token_var(block.ctx, op.loc)
-                new_loop_vars.results.append(res_var)
+                new_result_vars.append(res_var)
                 return body_var, res_var
 
             result_token_map = token_map.copy()
@@ -303,7 +265,7 @@ def _to_token_order_in_block(block: Block,
                     body_token_map[last_store_key], result_token_map[last_store_key] = \
                         append_new_carried_var(token_map[last_store_key])
                 else:
-                    raise ValueError(f"Unsupported memory effect: {effect}")
+                    raise TileInternalError(f"Unexpected memory effect: {effect}")
 
             if body_mem_effects.has_acquire_order:
                 body_token_map[ACQUIRE_TOKEN_KEY], result_token_map[ACQUIRE_TOKEN_KEY] = \
@@ -314,12 +276,13 @@ def _to_token_order_in_block(block: Block,
 
             _to_token_order_in_block(op.body, context, body_token_map,
                                      ifelse_info=None,
-                                     innermost_loop_info=InnermostLoopInfo(new_loop_vars,
-                                                                           body_mem_effects,
+                                     innermost_loop_info=InnermostLoopInfo(body_mem_effects,
                                                                            parallel_stores,
                                                                            token_map))
 
-            new_loop_op = Loop(op.body, op.loc, op.for_loop, new_loop_vars)
+            op.body.params = tuple(new_body_params)
+            new_loop_op = Loop(op.start, op.stop, op.step, tuple(new_initial_values),
+                               tuple(new_result_vars), op.body, op.loc)
             operations.append(new_loop_op)
 
             token_map = result_token_map
@@ -327,16 +290,13 @@ def _to_token_order_in_block(block: Block,
         elif isinstance(op, Continue):
             tokens = _get_cf_exit_tokens(innermost_loop_info.loop_mem_effects, token_map)
 
-            new_continue_op = Continue(op.loc, tuple(op.next_vars) + tokens,
-                                       innermost_loop_info.loop_vars,
-                                       op.is_for_loop_body)
+            new_continue_op = Continue(op.loc, tuple(op.values) + tokens)
             operations.append(new_continue_op)
 
         elif isinstance(op, Break):
             tokens = _get_cf_exit_tokens(innermost_loop_info.loop_mem_effects, token_map)
 
-            new_break_op = Break(op.loc, tuple(op.output_vars) + tokens,
-                                 innermost_loop_info.loop_vars)
+            new_break_op = Break(op.loc, tuple(op.values) + tokens)
             operations.append(new_break_op)
 
         elif isinstance(op, IfElse):
@@ -346,12 +306,11 @@ def _to_token_order_in_block(block: Block,
             merged_mem_effects = then_mem_effects | else_mem_effects
 
             result_token_map = token_map.copy()
-            new_res = IfElseResults(list(op.results.names), list(op.results.vars))
+            new_result_vars = list(op.result_vars)
 
             def add_new_ifelse_result():
                 x = _make_token_var(block.ctx, op.loc)
-                new_res.names.append(x.name)
-                new_res.vars.append(x)
+                new_result_vars.append(x)
                 return x
 
             for alias_set, effect in merged_mem_effects.items():
@@ -366,7 +325,7 @@ def _to_token_order_in_block(block: Block,
                     result_token_map[last_op_key] = add_new_ifelse_result()
                     result_token_map[last_store_key] = add_new_ifelse_result()
                 else:
-                    raise ValueError(f"Unsupported memory effect: {effect}")
+                    raise TileInternalError(f"Unexpected memory effect: {effect}")
 
             if merged_mem_effects.has_acquire_order:
                 result_token_map[ACQUIRE_TOKEN_KEY] = add_new_ifelse_result()
@@ -375,9 +334,9 @@ def _to_token_order_in_block(block: Block,
             for nested_block in op.nested_blocks:
                 _to_token_order_in_block(nested_block, context, token_map.copy(),
                                          innermost_loop_info=innermost_loop_info,
-                                         ifelse_info=IfElseInfo(new_res, merged_mem_effects))
+                                         ifelse_info=IfElseInfo(merged_mem_effects))
 
-            new_ifelse_op = IfElse(op.cond, op.then_block, op.else_block, op.loc, new_res)
+            new_ifelse_op = IfElse(op.cond, op.then_block, op.else_block, new_result_vars, op.loc)
             operations.append(new_ifelse_op)
 
             token_map = result_token_map
@@ -385,8 +344,7 @@ def _to_token_order_in_block(block: Block,
         elif isinstance(op, EndBranch):
             tokens = _get_cf_exit_tokens(ifelse_info.ifelse_mem_effects, token_map)
 
-            new_end_branch_op = EndBranch(op.loc, tuple(op.outputs) + tokens,
-                                          ifelse_info.ifelse_results)
+            new_end_branch_op = EndBranch(op.loc, tuple(op.outputs) + tokens)
             operations.append(new_end_branch_op)
 
         else:
@@ -451,9 +409,7 @@ def _get_input_token(token_key: TokenKey,
     return ret_tok, ret_op
 
 
-def _to_token_ordered_mem_op(op: MemoryOperation,
-                             token: Var,
-                             result_token: Var) -> Operation:
+def _to_token_ordered_mem_op(op, token: Var, result_token: Var) -> Operation:
 
     new_class = _TOKEN_ORDERED_OP_MAP[op.__class__]
     new_kwargs = dict(op.operands)
@@ -502,7 +458,7 @@ def _get_parallel_stores(
     Common in LayerNorm and RMSNorm patterns.
     """
 
-    if loop_op.for_loop is None:
+    if not loop_op.is_for_loop:
         return set()
 
     # Skips this optimization if alias_set size > 1 is present in the loop body
@@ -533,7 +489,7 @@ def _get_parallel_stores(
         tile_store_candidates.add(mem_ops[0])
 
     # Filter in stores that have non-overlapping indices
-    res = _filter_by_store_index(loop_op, tile_store_candidates, context.var_info)
+    res = _filter_by_store_index(loop_op, tile_store_candidates)
     return res
 
 
@@ -549,40 +505,15 @@ def _get_nested_mem_effects(
 
 
 def _filter_by_store_index(loop_op: Loop,
-                           tile_store_candidates: Set[Operation],
-                           var_info: VarInfo) -> Set[Operation]:
+                           tile_store_candidates: Set[Operation]) -> Set[Operation]:
 
     def is_idx_injective(idx_var: Var) -> bool:
-        root_idx_var = var_info.root_var.get(idx_var.name, idx_var.name)
-        if loop_op.for_loop and root_idx_var == loop_op.for_loop.induction_var.name:
-            iterable = loop_op.for_loop.iterable.name
-            iterable = var_info.root_var.get(iterable, iterable)
-            return isinstance(var_info.defining_op.get(iterable), Range)
         # TODO: allow more complex injective check: j = i * 2 + 3
-        return False
+        return loop_op.is_for_loop and idx_var.name == loop_op.induction_var.name
 
-    res = set()
-    for store_op in tile_store_candidates:
-        # TODO: support optimization on non-contiguous arrays
-        if not _get_input_var(store_op).get_type().elements_disjoint:
-            continue
-        # unpack index_var
-        index_ty = store_op.index.get_type()
-        if isinstance(index_ty, TupleTy):
-            root_var = var_info.root_var.get(store_op.index.name, store_op.index.name)
-            try:
-                # TODO: is it guanratee that the root_var has a defining op?
-                tuple_op = var_info.defining_op[root_var]
-                # TODO: handle case where tuple does not come from BuildTuple
-                assert isinstance(tuple_op, BuildTuple)
-            except (KeyError, AssertionError):
-                continue
-            unpacked = tuple_op.items
-        else:
-            unpacked = (store_op.index,)
-        if any(is_idx_injective(idx_var) for idx_var in unpacked):
-            res.add(store_op)
-    return res
+    return set(store_op for store_op in tile_store_candidates
+               if _get_input_var(store_op).get_type().elements_disjoint
+               and any(is_idx_injective(idx_var) for idx_var in store_op.index))
 
 
 def _try_loop_parallel_store(
