@@ -181,6 +181,38 @@ class Array:
             int (constant):
         """
 
+    @function
+    def slice(self, axis, start, stop) -> "Array":
+        """Creates a view of the |array| sliced along a single `axis`.
+
+        The returned array references the same underlying memory as |array|,
+        but with a restricted range from index `start` (inclusive) to `stop` (exclusive)
+        along the specified axis. No data is copied.
+
+        `axis` must be a constant integer. Negative values are supported and count
+        from the last dimension (e.g., ``axis=-1`` refers to the last axis).
+
+        `start` and `stop` must be integers (scalars or 0D tiles).
+        They must satisfy ``0 <= start < N`` and ``start <= stop <= N``, where ``N``
+        is the size of `array` along the sliced axis.
+
+        For example, consider a 2-dimensional array A of shape ``(M, N)``.
+        Slicing along axis 0 from `start` to `stop`:
+
+            >>> sub = A.slice(axis=0, start=start, stop=stop)
+
+        The result `sub` will be an array of shape ``(stop - start, N)``.
+        Using NumPy slice notation for illustration, this is equivalent to::
+
+            sub = A[start:stop, :]  # NumPy notation for reference only
+
+        The slice bounds can be dynamic (runtime values):
+
+            >>> # Process variable-length segments
+            >>> segment = A.slice(axis=1, start=offset, stop=offset + length)
+            >>> tile = ct.load(segment, (0, 0), shape=(TILE_M, TILE_N))
+        """
+
 
 class Tile:
     """A *tile array* (or *tile*) is an immutable multidimensional collection of values that is
@@ -1310,6 +1342,31 @@ def argmin(x, /, axis=None, *, keepdims=False) -> Tile:
     pass
 
 
+@function
+def reduce(x, /, axis, func, identity, *, keepdims=False):
+    """
+    Apply custom reduction function along axis.
+
+    Args:
+        x: input tile or a tuple of tiles to be reduced. If a tuple is provided, shapes
+            of the tiles in the tuple must be broadcastable to a common shape.
+        axis (int): an integer constant that specifies the axis to reduce along.
+        func: function for combining two values. If `x` is a single tile, then the function
+            must take two 0d tile arguments and return the combined 0d tile. For example,
+            `lambda a, b: a + b` or `operator.add` can be used to implement the sum reduction.
+            If `x` is a tuple of N tiles, then the function takes 2N tiles and returns a tuple
+            of N combined tiles. The first N arguments correspond to one of the groups of values
+            being combined, while the rest correspond to the other.
+        identity: a constant scalar or a tuple of constant scalars that specifies the identity
+            element of the `func`.
+        keepdims (bool): True to remove the reduced axis, False to keep the axis of size 1.
+            Default: False.
+
+    Returns:
+        Reduced tile, or tuple of reduced tiles, depending on the type of `x`.
+    """
+
+
 # ======== Scan ==============
 def _doc_scan_op(f):
 
@@ -1755,6 +1812,12 @@ def ceil(x, /) -> TileOrScalar:
     pass
 
 
+@_doc_unary_op
+@function
+def abs(x, /) -> TileOrScalar:
+    pass
+
+
 @function
 def negative(x, /) -> TileOrScalar:
     """Same as `-x`.
@@ -1891,10 +1954,8 @@ def assert_(cond, /, message=None) -> None:
 
 # ==== Private stubs ====
 
+
+def _m_array_slice(array, axis, start, stop): ...  # Array.slice(axis, start, stop)
+
+
 def _m_tile_item(tile): ...  # Tile.item()
-
-
-def _build_tuple(*items): ...
-
-
-def _identity(x): ...   # Identity function (i.e. returns x)
